
import json
import time
import datetime
import re

import tqdm
import pandas as pd

from hashlib import md5
from typing import Iterable
from functools import reduce
from typing import Any
from typing import List

from websocket import create_connection
from termcolor import colored
from .hf_server.cli_utils import R, ModelCard, ModelList, ModelCallResponse,ModelCallRequest,ClearRequest
from .hf_server.cli_utils import MachineInfo


import requests
class SmallModel:
    remote_host_ :str = None
    def __init__(self,name, remote=None):
        if remote is None:
            remote = self.__class__.remote_host_
        assert remote is not None
        self.remote_host  = remote
        self.name = name
        self._ok = False
        self.purpose = "unKnow"
    
    
    def status(self):
        name = self.name
        if "/" in name:
            name = name.rsplit("/",1)[-1]
        for i in self.__class__.show_all_loaded_models():
            if i.id == name:
                return True
        return False
    
    @classmethod
    def from_remote(cls, name,remote=None):
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        model = cls(name, remote)
        model.try_load_in_remote()
        return model
    
    @classmethod
    def gpu_info_remote(cls, remote=None) ->MachineInfo:
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        url = f"http://{remote}:15001/v1/device/info"
        return MachineInfo.parse_obj(next(R(url, method="get")))
    
    @classmethod
    def clean_all(cls, remote=None):
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        all_models = cls.show_all_loaded_models(remote)
        url = f"http://{remote}:15001/v1/models/clear"
        o = ClearRequest(ids=[i.id for i in all_models])
        return next(R(url, object=o))
        
    
    @classmethod
    def show_all_loaded_models(cls, remote=None)->List[ModelCard]:
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        url = f"http://{remote}:15001/v1/device/info"
        return MachineInfo.parse_obj(next(R(url, method="get"))).loaded_models.data
        

    @classmethod
    def show_all_models(cls, remote=None) -> List[ModelCard]:
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        url = f"http://{remote}:15001/v1/models"
        return ModelList.parse_obj(next(R(url, method="get"))).data
        
    def clean(self):
        try:
            # ws = create_connection(f"ws://{self.remote_host}:15000")
            # user_id = md5(time.asctime().encode()).hexdigest()
            # TODAY = datetime.datetime.now()
            # PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
            # ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
            # res = ws.recv()
            # if res != "ok":
            #     print(colored("[info]:","yellow") ,res)
            #     raise Exception("password error")
            # res = self.send_and_recv(json.dumps({"embed_documents":[self.name], "method":"clean"}),ws)
            # return res["embed"]
        
            name = self.name
            if "/" in name:
                name = name.rsplit("/",1)[-1]
            url = f"http://{self.remote_host}:15001//v1/models/clear"
            o = ClearRequest(ids=[name])
            m = MachineInfo.parse_obj(next(R(url, object=o)))
            self._ok = False
            for i in  m.loaded_models.data:
                if i.id == name:
                    self._ok = True
        except Exception as e:
            raise e
        finally:
            pass

    
    # def down_remote(self, try_time=3):
    #     try:
    #         ws = create_connection(f"ws://{self.remote_host}:15000")
    #         user_id = md5(time.asctime().encode()).hexdigest()
    #         TODAY = datetime.datetime.now()
    #         PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
    #         ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
    #         res = ws.recv()
    #         if res != "ok":
    #             print(colored("[info]:","yellow") ,res)
    #             raise Exception("password error")
    #         res = self.send_and_recv(json.dumps({"embed_documents":[self.name], "method":"clone"}),ws)
    #         time.sleep(2)
    #         res = self.msg()
    #         if "git clone failed: exit status 128" in res:
    #             print(colored("[Err : ]:","yellow") , colored(res,"red"))
    #             if try_time > 0:
    #                 return self.down_remote(try_time-1)
    #         return res["embed"]
            
    #     except Exception as e:
    #         raise e
    #     finally:
    #         ws.close()

    # @classmethod
    # def send_and_recv(cls, data, ws):
    #     try:
    #         T = len(data)// (1024*102)
    #         bart = tqdm.tqdm(total=T,desc=colored(" + sending data","cyan"))
    #         bart.leave = False
    #         for i in range(0, len(data), 1024*102):
    #             bart.update(1)
    #             ws.send(data[i:i+1024*102])
    #         bart.clear()
    #         bart.close()

    #         ws.send("[STOP]")
    #         message = ""
    #         total = int(ws.recv())
    #         bar = tqdm.tqdm(desc=colored(" + receiving data","cyan", attrs=["bold"]), total=total)
    #         bar.leave = False
    #         while 1:
    #             res = ws.recv()
    #             message += res
    #             bar.update(len(res))
    #             if message.endswith("[STOP]"):
    #                 message = message[:-6]
    #                 break
    #         bar.clear()
    #         bar.close()
    #         msg = json.loads(message)
    #         return msg
    #     except Exception as e:
    #         raise e
    
    # def msg(self):
    #     try:
    #         ws = create_connection(f"ws://{self.remote_host}:15000")
    #         user_id = md5(time.asctime().encode()).hexdigest()
    #         TODAY = datetime.datetime.now()
    #         PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
    #         ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
    #         res = ws.recv()
    #         if res != "ok":
    #             print(colored("[info]:","yellow") ,res)
    #             raise Exception("password error")
    #         res = self.send_and_recv(json.dumps({"embed_documents":[self.name], "method":"msg"}),ws)
    #         return res.get("embed","")
    #     except Exception as e:
    #         raise e
    #     finally:
    #         ws.close()
    
    # def change_remote_name(self, new_name):
    #     assert "/" not in new_name
    #     assert " " not in new_name
    #     ss = re.findall(r"[\w\-\_]+",new_name)
    #     assert len(ss) == 1 and ss[0] == new_name
    #     if self.name in self.show_all_models():
    #         try:
    #             ws = create_connection(f"ws://{self.remote_host}:15000")
    #             user_id = md5(time.asctime().encode()).hexdigest()
    #             TODAY = datetime.datetime.now()
    #             PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
    #             ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
    #             res = ws.recv()
    #             if res != "ok":
    #                 print(colored("[info]:","yellow") ,res)
    #                 raise Exception("password error")
    #             res = self.send_and_recv(json.dumps({"embed_documents":[self.name, new_name], "method":"change_name"}),ws)
    #             return res["embed"]
    #         except Exception as e:
    #             raise e
    #         finally:
    #             ws.close()
    #     else:
    #         raise Exception("model not exists")

    def check(self):
        self._ok = False
        name = self.name
        if "/" in name:
            name = name.rsplit("/",1)[-1]
        self.try_load_in_remote()
        return self._ok
    
    def try_load_in_remote(self):
        try:
            self._ok = False
            name = self.name
            if "/" in name:
                name = name.rsplit("/",1)[-1]
            url = f"http://{self.remote_host}:15001/v1/models/load"
            
            for i in ModelList.parse_obj(next(R(url, object=ModelCard(id=name)))).data:
                if i.id == name:
                    self._ok = True
                    return True
                
            return False
        except Exception as e:
            raise e
            return False

    # def show_remote_models(self):
    #     try:
    #         name = self.name
    #         if "/" in name:
    #             name = name.rsplit("/",1)[-1]
                
    #         ws = create_connection(f"ws://{self.remote_host}:15000")
    #         user_id = md5(time.asctime().encode()).hexdigest()
    #         TODAY = datetime.datetime.now()
    #         PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
    #         ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
    #         res = ws.recv()
    #         if res != "ok":
    #             print(colored("[info]:","yellow") ,res)
    #             raise Exception("password error")
    #         return self.send_and_recv(json.dumps({"embed_documents":[name], "method":"show"}), ws)["embed"]
    #     except Exception as e:
    #         raise e
    #     finally:
    #         ws.close()

    def __call__(self, args: List[str], pandas=False) -> Any:
        if isinstance(args, str):
            args = [args]
        assert isinstance(args, (list, tuple,Iterable,))
        if not self._ok:
            self.check()

        if not self._ok :
            raise Exception("remote's service no such model deployed"+self.name)
        name = self.name
        if "/" in name:
            name = name.rsplit("/",1)[-1]
        url = f"http://{self.remote_host}:15001/v1/models/call"
            
        try:
            o = ModelCallRequest(id=name, messages=args)

            no = 0
            oss = []
            for res in R(url, object=o, use_stream=True):
                ri = ModelCallResponse.parse_obj(res)
                if isinstance(ri.data, List):
                    if  len(ri.data) > 0 and isinstance(ri.data[0], str):
                        oss += ri.data
                    elif isinstance(ri.data[0], dict):
                        for ii in ri.data:
                            ar = args[no]
                            ii["input"] = ar
                            no += 1
                            oss.append(ii)
                    else:
                        oss += ri.data
                    
            if len(oss) > 0 and isinstance(oss[0], dict) and pandas:
                return pd.DataFrame(oss)
            return oss
            
        except Exception as e:
            raise e
            # import ipdb;ipdb.set_trace()
    
    def _merged(self, res_list):
        res = []
        for one_obj in res_list:
            if isinstance(one_obj, list) :
                if len(one_obj) > 1:
                    if isinstance(one_obj[0], dict):
                        res.append({k:reduce(lambda x,y: x+y , map(lambda x: x[k] ,one_obj)) for k in one_obj[0]})
                    else:
                        res.append(one_obj)
                else:
                    res.append(one_obj[0])
            else:
                res.append(one_obj)
        return res




# class SmallModel_v2(SmallModel):

#     def status(self):
#         return super().status()

#     def try_load_in_remote(self):
#         def try_load_in_remote(self):
#         try:
#             self._ok = False
#             name = self.name
#             if "/" in name:
#                 name = name.rsplit("/",1)[-1]
            
            


from langchain.llms.base import LLM
from typing import Iterable, List
import re
from typing import List, Dict, Any, Optional, Union
import pathlib
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from .callbacks import AsyncWebsocketHandler, AsyncWebSocksetCallbackManager
import langchain
try:
    from langchain.cache import GPTCache
except:
    pass
import gptcache
from gptcache.processor.pre import get_prompt
from gptcache.manager.factory import get_data_manager
from aiowebsocket.converses import AioWebSocket
from websocket import create_connection
import websockets
from websockets.server import serve
from langchain.embeddings import HuggingFaceEmbeddings
from termcolor import colored
import datetime
from hashlib import md5
import time
import json
import inspect

try:
    
    from transformers import AutoTokenizer, AutoModel, AutoModelForSequenceClassification
    from fastchat.serve.inference import load_model
    from fastchat.conversation import get_default_conv_template
    from accelerate import load_checkpoint_and_dispatch
    
    from fastchat.serve.inference import generate_stream
    import torch, gc
    from .hf_server.cli_utils import create_chat_completion, ChatMessage, acreate_chat_completion
    DEFAULT_CACHE_MAP_PATH = str(pathlib.Path.home() / ".cache" / "local_qa_cache_map")
    i = 0

    def init_gptcache_map(cache_obj: gptcache.Cache):
        global i
        cache_path = f'{DEFAULT_CACHE_MAP_PATH}_{i}.txt'
        cache_obj.init(
            pre_embedding_func=get_prompt,
            data_manager=get_data_manager(data_path=cache_path),
        )
        i += 1
    
    langchain.llm_cache = GPTCache(init_gptcache_map)
    # print(colored("init gptcache", "green"))
except Exception as e:
    # raise e
    print("use remote ignore this / load transformers failed, please install transformers and accelerate first and torch.")

def is_async_method(method):
    return inspect.iscoroutinefunction(method)

def load_model_on_gpus_old(checkpoint_path, num_gpus: int = 2,device_map = None, **kwargs):
    if num_gpus < 2 and device_map is None:
        model = AutoModel.from_pretrained(checkpoint_path, trust_remote_code=True, **kwargs).half().cuda()
    else:
        from accelerate import dispatch_model
        model = AutoModel.from_pretrained(checkpoint_path, trust_remote_code=True, **kwargs).half()
        if device_map is None:
            device_map = auto_configure_device_map(num_gpus)
        model = dispatch_model(model, device_map=device_map)

    return model
def auto_configure_device_map(num_gpus: int) -> Dict[str, int]:
    # transformer.word_embeddings 占用1层
    # transformer.final_layernorm 和 lm_head 占用1层
    # transformer.layers 占用 28 层
    # 总共30层分配到num_gpus张卡上
    num_trans_layers = 28
    per_gpu_layers = 30 / num_gpus

    # bugfix: 在linux中调用torch.embedding传入的weight,input不在同一device上,导致RuntimeError
    # windows下 model.device 会被设置成 transformer.word_embeddings.device
    # linux下 model.device 会被设置成 lm_head.device
    # 在调用chat或者stream_chat时,input_ids会被放到model.device上
    # 如果transformer.word_embeddings.device和model.device不同,则会导致RuntimeError
    # 因此这里将transformer.word_embeddings,transformer.final_layernorm,lm_head都放到第一张卡上
    device_map = {'transformer.word_embeddings': 0,
                  'transformer.final_layernorm': 0, 'lm_head': 0}

    used = 2
    gpu_target = 0
    for i in range(num_trans_layers):
        if used >= per_gpu_layers:
            gpu_target += 1
            used = 0
        assert gpu_target < num_gpus
        device_map[f'transformer.layers.{i}'] = gpu_target
        used += 1

    return device_map

def enforce_stop_tokens(text: str, stop: List[str]) -> str:
    """Cut off the text as soon as any stop words occur."""
    return re.split("|".join(stop), text)[0]


def auto_gc():
    if torch.cuda.is_available():
        # for all cuda device:
        for i in range(0,torch.cuda.device_count()):
            CUDA_DEVICE = f"cuda:{i}"
            with torch.cuda.device(CUDA_DEVICE):
                torch.cuda.empty_cache()
                torch.cuda.ipc_collect()
    else:
        gc.collect()

class ChatGLMLLM(LLM):
    """
            Load a model from local or remote
        if want to use stream mode:
            'streaming=True'
        if want to use langchain's Callback:
            examples: 'callbacks=[StreamingStdOutCallbackHandler(), AsyncWebsocketHandler()]'

        if want use cpu: # default will try to use gpu
            'cpu=True'
        
        if want to use remote's model:
            'remote_host="xx.xxx.xx.xx"'  , if set this , will auto call by ws://xx.xxx.xxx.xx:15000"
            optional:
                remote_callback: a callback function, will call when receive a new token like  'callback(new_token, history, response)'
                if not set, will print to stdout

    """
    max_tokens: int = 10000
    temperature: float = 0.01
    top_p = 0.9
    history = []
    history_id = "default"
    tokenizer: Any = None
    model: Any = None
    history_len: int = 10
    model: Any = None
    tokenizer: Any = None
    model_path: str = pathlib.Path.home() / ".cache" / "chatglm"
    cpu: bool = False
    streaming: bool = False
    verbose: bool = False
    callbacks :Any  = [StreamingStdOutCallbackHandler(), AsyncWebsocketHandler()]
    # callback_manager:  Any = BaseCallbackManager
    remote_host: Any = None

    def set_history(self, hist:List[str]):
        self.history = hist
    
    
    @property
    def _llm_type(self) -> str:
        return "ChatGLM"

    async def _acall(self, prompt: str, stop: List[str] = None):
        
        if self.remote_host is not None :
            uri = f"http://{self.remote_host}:15001/v1/chat/completions"
            result = ''
            msgs = [
                ChatMessage(role="system", content="You are ChatGLM3, a helpful assistant. Follow the user's instructions carefully. Respond using markdown.")
            ]
            for h in self.history:
                p,r = h
                msgs.append(ChatMessage(role="user", content=p.strip()))
                msgs.append(ChatMessage(role="assistant", content=r.strip()))
            msgs.append(ChatMessage(role="user", content=prompt))
            for call in self.callbacks:
                if is_async_method(call.on_llm_start):
                    await call.on_llm_start(prompt, None, verbose=self.verbose)
            result = ""
            ss = ""
            async for r in acreate_chat_completion(uri, msgs, temperature=self.temperature, max_tokens=self.max_tokens, top_p=self.top_p):
                for choice in r.choices:
                    if choice.delta.content is not None:
                        msg = {}
                        # yield choice.delta
                        msg["new"] = choice.delta.content
                        ss += choice.delta.content
                        msg["response"] = ss
                        msg["verbose"] = self.verbose
                        result = ss
                        for call in self.callbacks:
                            if is_async_method(call.on_llm_new_token):
                                data = {"response": ss, "history": self.history,"query": prompt}
                                await call.on_llm_new_token(data, verbose=self.verbose, **data)

            self.history = self.history+[[prompt, result]]
            for call in self.callbacks:
                if is_async_method(call.on_llm_start):
                    await call.on_llm_end(result, verbose=self.verbose)
            return result


    def stream(self,prompt: str, stop: List[str] = None):
        uri = f"http://{self.remote_host}:15001/v1/chat/completions"
        result = ''
        msgs = [
            ChatMessage(role="system", content="You are ChatGLM3, a helpful assistant. Follow the user's instructions carefully. Respond using markdown.")
        ]
        for h in self.history:
            p,r = h
            msgs.append(ChatMessage(role="user", content=p.strip()))
            msgs.append(ChatMessage(role="assistant", content=r.strip()))
        msgs.append(ChatMessage(role="user", content=prompt))

        gen = create_chat_completion(uri, msgs, temperature=self.temperature, max_tokens=self.max_tokens, top_p=self.top_p)
        ss = ""
        for r in gen:
            
            for choice in r.choices:
                if choice.delta.content is not None:
                    msg = {}
                    # yield choice.delta
                    msg["new"] = choice.delta.content
                    ss += choice.delta.content
                    msg["response"] = ss
                    msg["verbose"] = self.verbose
                    result = ss
                    yield msg

        self.history = self.history+[[prompt, result]]


    
    def _call(self, prompt: str, stop: List[str]  = None,run_manager: Any = None) -> str:
        
        if self.remote_host is not None :
            uri = f"http://{self.remote_host}:15001/v1/chat/completions"
            result = ''
            msgs = [
                ChatMessage(role="system", content="You are ChatGLM3, a helpful assistant. Follow the user's instructions carefully. Respond using markdown.")
            ]
            for h in self.history:
                p,r = h
                msgs.append(ChatMessage(role="user", content=p.strip()))
                msgs.append(ChatMessage(role="assistant", content=r.strip()))
            msgs.append(ChatMessage(role="user", content=prompt))

            gen = create_chat_completion(uri, msgs, temperature=self.temperature, max_tokens=self.max_tokens, top_p=self.top_p)
            ss = ""
            for callback in self.callbacks:
                if is_async_method(callback.on_llm_start):continue
                callback.on_llm_start(
                    None,
                    prompt,
                    run_id=0,
                    verbose=self.verbose
                )

            for r in gen:
                
                for choice in r.choices:
                    if choice.delta.content is not None:
                        msg = {}
                        # yield choice.delta
                        msg["new"] = choice.delta.content
                        new_token = choice.delta.content
                        ss += choice.delta.content
                        msg["response"] = ss
                        msg["verbose"] = self.verbose
                        result = ss
                        for callback in self.callbacks:
                            if is_async_method(callback.on_llm_start):continue
                            callback.on_llm_new_token(
                                new_token,
                                **msg
                            )
                        result = ss
            
            
            
            for callback in self.callbacks:
                if is_async_method(callback.on_llm_start):continue
                callback.on_llm_end(result, verbose=self.verbose)
            self.history = self.history+[[prompt, result]]
            return result



import requests
import aiohttp
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
import json
import sys
import inspect
from typing import Any, List, Optional
from termcolor import colored


import langchain
from langchain.llms.base import LLM
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from .callbacks import AsyncWebsocketHandler, AsyncWebSocksetCallbackManager

def is_async_method(method):
    return inspect.iscoroutinefunction(method)

class VllmBase(LLM):
    n = 1
    use_beam_search = False
    top_p:float =1.0
    temperature:float = 0.0
    max_tokens:int = 8096
    history_len: int = 10
    streaming: bool = True
    verbose: bool = False
    callbacks :Any  = [StreamingStdOutCallbackHandler()]
    remote_host: str = None
    hist = []
    system = "You are a assistant."
    
        
    def clear_history(self):
        self.hist = []

    def send(self, prompt,stop=[], **kwargs):
        headers = {"User-Agent": "Test Client"}
        basepload = {
            "prompt": prompt,
            "n": self.n,
            "use_beam_search": self.use_beam_search,
            "top_p": self.top_p,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "stream": True,

        }
        if stop and len(stop) > 0:
            basepload["stop"] = stop
        basepload.update(kwargs)
        basepload = self.make_args(basepload)
        if self.verbose:
            print(prompt)
            print(basepload)
        api_url = f"https://{self.remote_host}:15443/generate"
        response = requests.post(api_url, headers=headers, 
                                 verify=False,
                                 json=basepload, stream=True)
        
        return response
    
    def make_args(self, parameter):
        raise NotImplementedError("must implement make_args(parameter)")
    
    async def asend(self, prompt,stop=[], **kwargs):
        """
        async send use aiohttp
        """
        headers = {"User-Agent": "Test Client"}
        basepload = {
            "prompt": prompt,
            "n": self.n,
            "use_beam_search": self.use_beam_search,
            "top_p": self.top_p,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            
            "stream": True,
        }
        if stop and len(stop) > 0:
            basepload["stop"] = stop
        basepload.update(kwargs)
        basepload = self.make_args(basepload)
        api_url = f"https://{self.remote_host}:15443/generate"
        async with aiohttp.ClientSession() as session:
            async with session.post(api_url, headers=headers, 
                                    verify_ssl=False,
                                    json=basepload, 
                                    timeout=aiohttp.ClientTimeout(total=60*60*24),
                                    ) as response:
                one = ""
                old_data = b""
                async for chunk in response.content.iter_chunked(8192):
                                    #  decode_unicode=False,
                                    #  delimiter=b"\0"):
                    if chunk:
                        if old_data != b"":
                            chunk = old_data + chunk
                            old_data = b""

                        if not chunk.endswith(b"\0"):
                            old_data = chunk.rsplit(b"\0",1)[1]
                            chunk = chunk.rsplit(b"\0",1)[0]

                        for line in chunk.split(b"\0"):
                            if line:
                                data = json.loads(line.decode("utf-8"))
                                output = data["text"]
                                for lineno, linetext in enumerate(output):
                                    linetext = self.parse_output(linetext)
                                    new_text = linetext[len(one):]
                                    one = linetext
                                    yield {
                                        "new": new_text,
                                        "history": self.hist,
                                        "lineno": lineno,
                                        "response": one,
                                    }

                self.hist.append((prompt, one))

    async def astream(self, prompt, history=[],stop=[],system="You are a assistent.", **kwargs):
        if len(history) > 0:
            query = self.make_context(prompt, history, system)
            hist = history
        else:
            query = self.make_context(prompt, self.hist, system)
            hist = self.hist
        
        async for data in  self.asend(query,stop=stop, **kwargs):
            yield data
        
        
    @property
    def _llm_type(self) -> str:
        return self.__class__.__name__

    async def _acall(self, prompt: str, stop: List[str] = None, run_manager: any = None,):
        
        query = self.make_context(prompt, self.hist, self.system)
        one = ""
        try:
            if hasattr(run_manager, "on_llm_start"):
                await run_manager.on_llm_start(
                    None,
                    prompt,
                    verbose=self.verbose
                )
            
            async for data in  self.asend(query,stop=stop):
                delta = data["new"]
                await run_manager.on_llm_new_token(
                    delta, verbose=self.verbose, **data
                )
                one = data["response"]
            await run_manager.on_llm_end(
                one, verbose=self.verbose
            )
            # self.hist.append((prompt, one))

            return one
        except Exception as e:
            if hasattr(run_manager, "on_llm_error"):
                await run_manager.on_llm_error(e)
            else:
                print(e)

    def _call(self, prompt: str, stop: List[str] = None, run_manager:any = None,):
        if stop is not None:
            self.stop = stop
        query = self.make_context(prompt, self.hist, self.system)
        response = self.send(query)
        one = ""
        try:
            if hasattr(run_manager, "on_llm_start"):
                run_manager.on_llm_start(
                    None,
                    prompt,
                    verbose=self.verbose
                )

            for chunk in response.iter_lines(chunk_size=8192,
                                        decode_unicode=False,
                                        delimiter=b"\0"):
                if chunk:
                    data = json.loads(chunk.decode("utf-8"))
                    output = data["text"]
                    for lineno, linetext in enumerate(output):
                        linetext = self.parse_output(linetext)
                        new_text = linetext[len(one):]
                        one = linetext
                        
                        data = {
                            "new": new_text,
                            "lineno": lineno,
                            "history": self.hist,
                            "response": one,
                        }

                        delta = new_text
                    
                        run_manager.on_llm_new_token(
                            delta,
                            **data
                        )
                
            self.hist.append((prompt, one))
        except Exception as e:
            if self.verbose:
                print(e)
            run_manager.on_llm_error(e)
        run_manager.on_llm_end(one, verbose=self.verbose)
        return one

    def stream(self, prompt, history=[],stop=[],system="You are a assistent.", **kwargs):
        if len(history) > 0:
            query = self.make_context(prompt, history, system)
            hist = history
        else:
            query = self.make_context(prompt, self.hist, system)
            hist = self.hist
        response = self.send(query, stop=stop, **kwargs)
        one = ""
        
        for chunk in response.iter_lines(chunk_size=8192,
                                     decode_unicode=False,
                                     delimiter=b"\0"):
            if chunk:
                data = json.loads(chunk.decode("utf-8"))
                output = data["text"]
                for lineno, linetext in enumerate(output):
                    linetext = self.parse_output(linetext)
                    new_text = linetext[len(one):]
                    one = linetext
                    
                    yield {
                        "new": new_text,
                        "lineno": lineno,
                        "history": hist,
                        "response": one,
                    }
        self.hist.append((prompt, one))
        
        
    def stream_stdout(self, prompt, history=[], stop=[], system="You are a assistent.", **kwargs):
        print(">>",colored(prompt, "green"), end="\nAssist:")
        for data in self.stream(prompt,stop=stop,history=history,system=system, **kwargs):
            sys.stdout.write(colored(data["new"], "blue"))
            sys.stdout.flush()
        print()

    def make_context(prompt, system="", history=[]):
        raise NotImplementedError("must implement make_context(prompt, history=[(q,h)])")

    def parse_output(self, output):
        raise NotImplementedError("must implement parse_output(output)")


class NormalLLM(VllmBase):
    start_token = ""
    end_token = ""

    def make_args(self, parameter):
        # parameter["stop"] = ['<|im_end|>','<|endoftext|>']
        # parameter["stop"] = '<|im_end>'
        return parameter

    def make_context(self, prompt, history=[],system="You are a assistant."):
        def _token(role, content):
            return f"{role}:\n{content}"

        raw_text = _token("system", system) + "\n"
        for q, h in history:
            raw_text += _token("user", q)
            raw_text += "\n"+_token("assistant", h) +"\n"
        return raw_text + _token("user", prompt) + "\n" + self.start_token + "assistant\n"

    def parse_output(self, output):
        return output

class QwenLLM(VllmBase):

    start_token = "<|im_start|>"
    end_token = "<|im_end|>"
    
    def make_args(self, parameter):
        parameter["stop"] = ['<|im_end|>','<|endoftext|>']
        # parameter["stop"] = '<|im_end>'
        return parameter

    def make_context(self, prompt, history=[],system="You are a assistant."):
        def _token(role, content):
            return f"{self.start_token}{role}\n{content}{self.end_token}"

        raw_text = _token("system", system) + "\n"
        for q, h in history:
            raw_text += _token("user", q)
            raw_text += "\n"+_token("assistant", h) +"\n"
        return raw_text + _token("user", prompt) + "\n" + self.start_token + "assistant\n"

    def parse_output(self, output):
        if self.start_token+"assistant" not in output:
            if output.endswith(self.end_token):
                return output.rsplit(self.end_token,1)[0]
            return output
        else:
            l = output.rsplit(self.start_token+"assistant",1)[1]
            if l.endswith(self.end_token):
                return l.rsplit(self.end_token,1)[0]
            return l


# if __name__ == "__main__":
#     import argparse

#     parser = argparse.ArgumentParser()
#     parser.add_argument("-H","--host", default="localhost")
#     # loop
#     parser.add_argument("--loop", action="store_true")
#     #  query
#     parser.add_argument("-q","--query",nargs="*", default="")
#     args = parser.parse_args()
#     if len(args.query) > 0 or args.loop:
#         qwen = Qwen(args.host)

#         if args.loop:
#             while True:
#                 q = input("You:")
#                 qwen.stream_stdout(q)
#         else:
#             qwen.stream_stdout(" ".join(args.query))



import os
import gc
import json
import torch
from torch.nn import Module

try:
    from transformers import PreTrainedModel, PreTrainedTokenizer
    from transformers import AutoModel
    from transformers.generation.logits_process import LogitsProcessor
except ImportError:
    pass
from typing import Dict, Union, Optional, Tuple


def auto_configure_device_map(num_gpus: int) -> Dict[str, int]:
    # transformer.word_embeddings 占用1层
    # transformer.final_layernorm 和 lm_head 占用1层
    # transformer.layers 占用 28 层
    # 总共30层分配到num_gpus张卡上
    num_trans_layers = 28
    per_gpu_layers = 30 / num_gpus

    # bugfix: 在linux中调用torch.embedding传入的weight,input不在同一device上,导致RuntimeError
    # windows下 model.device 会被设置成 transformer.word_embeddings.device
    # linux下 model.device 会被设置成 lm_head.device
    # 在调用chat或者stream_chat时,input_ids会被放到model.device上
    # 如果transformer.word_embeddings.device和model.device不同,则会导致RuntimeError
    # 因此这里将transformer.word_embeddings,transformer.final_layernorm,lm_head都放到第一张卡上
    # 本文件来源于https://github.com/THUDM/ChatGLM-6B/blob/main/utils.py
    # 仅此处做少许修改以支持ChatGLM3
    device_map = {
        'transformer.embedding.word_embeddings': 0,
        'transformer.encoder.final_layernorm': 0,
        'transformer.output_layer': 0,
        'transformer.rotary_pos_emb': 0,
        'lm_head': 0
    }

    used = 2
    gpu_target = 0
    for i in range(num_trans_layers):
        if used >= per_gpu_layers:
            gpu_target += 1
            used = 0
        assert gpu_target < num_gpus
        device_map[f'transformer.encoder.layers.{i}'] = gpu_target
        used += 1

    return device_map


def load_model_on_gpus(checkpoint_path: Union[str, os.PathLike], num_gpus: int = 2,
                       device_map: Optional[Dict[str, int]] = None, **kwargs) -> Module:
    if num_gpus < 2 and device_map is None:
        model = AutoModel.from_pretrained(checkpoint_path, trust_remote_code=True, **kwargs).half().cuda()
    else:
        from accelerate import dispatch_model

        model = AutoModel.from_pretrained(checkpoint_path, trust_remote_code=True, **kwargs).half()

        if device_map is None:
            device_map = auto_configure_device_map(num_gpus)

        model = dispatch_model(model, device_map=device_map)

    return model


class InvalidScoreLogitsProcessor(LogitsProcessor):
    def __call__(
            self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        if torch.isnan(scores).any() or torch.isinf(scores).any():
            scores.zero_()
            scores[..., 5] = 5e4
        return scores


def process_response(output: str, use_tool: bool = False) -> Union[str, dict]:
    content = ""
    for response in output.split("<|assistant|>"):
        metadata, content = response.split("\n", maxsplit=1)
        if not metadata.strip():
            content = content.strip()
            content = content.replace("[[训练时间]]", "2023年")
        else:
            if use_tool:
                content = "\n".join(content.split("\n")[1:-1])

                def tool_call(**kwargs):
                    return kwargs

                parameters = eval(content)
                content = {
                    "name": metadata.strip(),
                    "arguments": json.dumps(parameters, ensure_ascii=False)
                }
            else:
                content = {
                    "name": metadata.strip(),
                    "content": content
                }
    return content


@torch.inference_mode()
def generate_stream_chatglm3(model: PreTrainedModel, tokenizer: PreTrainedTokenizer, params: dict):
    messages = params["messages"]
    functions = params["functions"]
    temperature = float(params.get("temperature", 1.0))
    repetition_penalty = float(params.get("repetition_penalty", 1.0))
    top_p = float(params.get("top_p", 1.0))
    max_new_tokens = int(params.get("max_tokens", 256))
    echo = params.get("echo", True)
    messages = process_chatglm_messages(messages, functions=functions)
    query, role = messages[-1]["content"], messages[-1]["role"]

    inputs = tokenizer.build_chat_input(query, history=messages[:-1], role=role)
    inputs = inputs.to(model.device)
    input_echo_len = len(inputs["input_ids"][0])

    if input_echo_len >= model.config.seq_length:
        print(f"Input length larger than {model.config.seq_length}")

    eos_token_id = [
        tokenizer.eos_token_id,
        tokenizer.get_command("<|user|>"),
    ]

    gen_kwargs = {
        "max_new_tokens": max_new_tokens,
        "do_sample": True if temperature > 1e-5 else False,
        "top_p": top_p,
        "repetition_penalty": repetition_penalty,
        "logits_processor": [InvalidScoreLogitsProcessor()],
    }
    if temperature > 1e-5:
        gen_kwargs["temperature"] = temperature

    total_len = 0
    for total_ids in model.stream_generate(**inputs, eos_token_id=eos_token_id, **gen_kwargs):
        total_ids = total_ids.tolist()[0]
        total_len = len(total_ids)
        if echo:
            output_ids = total_ids[:-1]
        else:
            output_ids = total_ids[input_echo_len:-1]

        response = tokenizer.decode(output_ids)
        if response and response[-1] != "�":
            response, stop_found = apply_stopping_strings(response, ["<|observation|>"])

            yield {
                "text": response,
                "usage": {
                    "prompt_tokens": input_echo_len,
                    "completion_tokens": total_len - input_echo_len,
                    "total_tokens": total_len,
                },
                "finish_reason": "function_call" if stop_found else None,
            }

            if stop_found:
                break

    # Only last stream result contains finish_reason, we set finish_reason as stop
    ret = {
        "text": response,
        "usage": {
            "prompt_tokens": input_echo_len,
            "completion_tokens": total_len - input_echo_len,
            "total_tokens": total_len,
        },
        "finish_reason": "stop",
    }
    yield ret

    gc.collect()
    torch.cuda.empty_cache()


def process_chatglm_messages(messages, functions=None):
    _messages = messages
    messages = []
    if functions:
        messages.append(
            {
                "role": "system",
                "content": "Answer the following questions as best as you can. You have access to the following tools:",
                "tools": functions
            }
        )

    for m in _messages:
        role, content, func_call = m.role, m.content, m.function_call
        if role == "function":
            messages.append(
                {
                    "role": "observation",
                    "content": content
                }
            )

        elif role == "assistant" and func_call is not None:
            for response in content.split("<|assistant|>"):
                metadata, sub_content = response.split("\n", maxsplit=1)
                messages.append(
                    {
                        "role": role,
                        "metadata": metadata,
                        "content": sub_content.strip()
                    }
                )
        else:
            messages.append({"role": role, "content": content})
    return messages


def generate_chatglm3(model: PreTrainedModel, tokenizer: PreTrainedTokenizer, params: dict):
    for response in generate_stream_chatglm3(model, tokenizer, params):
        pass
    return response


def apply_stopping_strings(reply, stop_strings) -> Tuple[str, bool]:
    stop_found = False
    for string in stop_strings:
        idx = reply.find(string)
        if idx != -1:
            reply = reply[:idx]
            stop_found = True
            break

    if not stop_found:
        # If something like "\nYo" is generated just before "\nYou: is completed, trim it
        for string in stop_strings:
            for j in range(len(string) - 1, 0, -1):
                if reply[-j:] == string[:j]:
                    reply = reply[:-j]
                    break
            else:
                continue

            break

    return reply, stop_found



# from .llm import AsyncServer
from chatglm_src.serv import run
import uvicorn
import argparse
import pathlib

parser = argparse.ArgumentParser(description='Start a chatglm server.')
parser.add_argument('-p','--port', type=int, default=15001, help='port number')
parser.add_argument("-w",'--worker', type=int, default=1, help='worker number')
# parser.add_argument("-n","--name", default="chatglm", help="llm's name")
# parser.add_argument("-e","--extend", nargs="*", help="load extend models , like 'emotion' 'ner' in ~/.cache/ ")


def main():
    args = parser.parse_args()
    # if args.extend:
    #     for i in args.extend:
    #         AsyncServer.load_model(i)
    # AsyncServer.start(port=args.port, model_path=args.model_path, name=args.name)
    # uvicorn.run(app, host="0.0.0.0", port=args.port, workers=args.worker)
    
    run(port=args.port, workers=args.worker)


import os
from typing import Dict, Tuple, Union, Optional

from torch.nn import Module
from transformers import AutoModel


def auto_configure_device_map(num_gpus: int) -> Dict[str, int]:
    # transformer.word_embeddings 占用1层
    # transformer.final_layernorm 和 lm_head 占用1层
    # transformer.layers 占用 28 层
    # 总共30层分配到num_gpus张卡上
    num_trans_layers = 28
    per_gpu_layers = 30 / num_gpus

    # bugfix: 在linux中调用torch.embedding传入的weight,input不在同一device上,导致RuntimeError
    # windows下 model.device 会被设置成 transformer.word_embeddings.device
    # linux下 model.device 会被设置成 lm_head.device
    # 在调用chat或者stream_chat时,input_ids会被放到model.device上
    # 如果transformer.word_embeddings.device和model.device不同,则会导致RuntimeError
    # 因此这里将transformer.word_embeddings,transformer.final_layernorm,lm_head都放到第一张卡上
    # 本文件来源于https://github.com/THUDM/ChatGLM-6B/blob/main/utils.py
    # 仅此处做少许修改以支持ChatGLM2
    device_map = {
        'transformer.embedding.word_embeddings': 0,
        'transformer.encoder.final_layernorm': 0,
        'transformer.output_layer': 0,
        'transformer.rotary_pos_emb': 0,
        'lm_head': 0
    }

    used = 2
    gpu_target = 0
    for i in range(num_trans_layers):
        if used >= per_gpu_layers:
            gpu_target += 1
            used = 0
        assert gpu_target < num_gpus
        device_map[f'transformer.encoder.layers.{i}'] = gpu_target
        used += 1

    return device_map


def load_model_on_gpus(checkpoint_path: Union[str, os.PathLike], num_gpus: int = 2,
                       device_map: Optional[Dict[str, int]] = None, **kwargs) -> Module:
    if num_gpus < 2 and device_map is None:
        model = AutoModel.from_pretrained(checkpoint_path, trust_remote_code=True, **kwargs).half().cuda()
    else:
        from accelerate import dispatch_model

        model = AutoModel.from_pretrained(checkpoint_path, trust_remote_code=True, **kwargs).half()

        if device_map is None:
            device_map = auto_configure_device_map(num_gpus)

        model = dispatch_model(model, device_map=device_map)

    return model


from typing import List
from langchain.embeddings.base import Embeddings
from langchain.embeddings import HuggingFaceEmbeddings
from websocket import create_connection
from hashlib import md5
import time
import json
import tqdm
import datetime
from typing import List, Dict, Any, Optional, Union
from termcolor import colored

from .hf_server.cli_utils import R, MessageRequest


class ChatGLMEmbeding(Embeddings):

    remote_host :str = None

    def __init__(self, model_path: str=None, remote_host:str=None):
        if remote_host is None:
            self._emb = HuggingFaceEmbeddings(model_name=model_path)
        else:
            self.remote_host = remote_host

    # def send_and_recv(self, data, ws):
    #     try:
    #         T = len(data)// (1024*102)
    #         bart = tqdm.tqdm(total=T,desc=colored(" + sending data ","cyan"))
    #         bart.leave = False
    #         for i in range(0, len(data), 1024*102):
    #             bart.update(1)
    #             ws.send(data[i:i+1024*102])
    #         bart.clear()
    #         bart.close()

    #         ws.send("[STOP]")
    #         message = ""
    #         total = int(ws.recv())
    #         bar = tqdm.tqdm(desc=colored(" + receiving data","cyan", attrs=["bold"]), total=total)
    #         bar.leave = False
    #         while 1:
    #             res = ws.recv()
    #             message += res
    #             bar.update(len(res))
    #             if message.endswith("[STOP]"):
    #                 message = message[:-6]
    #                 break
    #         bar.clear()
    #         bar.close()
    #         msg = json.loads(message)
    #         return msg
    #     except Exception as e:
    #         raise e

    def embed_documents_remote(self, texts):
        # ws = create_connection(f"ws://{self.remote_host}:15000")
        uri = f"http://{self.remote_host}:15001/v1/embeddings"
        user_id = md5(time.asctime().encode()).hexdigest()
        # TODAY = datetime.datetime.now()
        # PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        # ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        # # time.sleep(0.5)
        # res = ws.recv()
        # if res != "ok":
        #     print(colored("[info]:","yellow") ,res)
        #     raise Exception("password error")
        
        # data = json.dumps({"embed_documents":texts})
        try:
            m = MessageRequest(messages=texts, stream=True)
            ss = []
            for i in R(uri, m, use_stream=True):
                ss += i["data"]
            return ss
        except Exception as e:
            print(e)
            import ipdb;ipdb.set_trace()
    
    def embed_query_remote(self, text):
        uri = f"http://{self.remote_host}:15001/v1/embeddings"
        user_id = md5(time.asctime().encode()).hexdigest()
        # TODAY = datetime.datetime.now()
        # PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        # ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        # # time.sleep(0.5)
        # res = ws.recv()
        # if res != "ok":
        #     print(colored("[info]:","yellow") ,res)
        #     raise Exception("password error")
        
        # data = json.dumps({"embed_documents":texts})
        try:
            m = MessageRequest(messages=[text], stream=True)
            ss = []
            for i in R(uri, m, use_stream=True):
                ss += i["data"]
            return ss[0]
        except Exception as e:
            print(e)
            import ipdb;ipdb.set_trace()

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        if self.remote_host is not None:
            return self.embed_documents_remote(texts)
        return self._emb.embed_documents(texts)

    def embed_query(self, text: str) -> List[float]:
        if self.remote_host is not None:
            return self.embed_query_remote(text)
        return self._emb.embed_query(text)


import pathlib
from .llm import ChatGLMLLM,Vicuna13B
from .qwen import QwenLLM
from .chains import KnowdageQA
from .embeding import ChatGLMEmbeding as Embeddings
from .cluster_and_smi import papers as QueryTexts
from .baichuan import Baichuan2LLM

# _EMBEDDING_PATH = pathlib.Path.home() / ".cache" / "chatglm-embedding"

# if not _EMBEDDING_PATH.exists():
#     print("------ start founding embedding  ------:", _EMBEDDING_PATH)
#     import os
#     # unzip emb.zip to _EMBEDDING_PATH
#     import zipfile
    
#     os.makedirs(str(pathlib.Path.home() / ".cache"),exist_ok=True)
    
#     with zipfile.ZipFile(str(pathlib.Path(__file__).parent /"Res"/ "emb.zip"), 'r') as zip_ref:
#         zip_ref.extractall(str(_EMBEDDING_PATH))
#         print("------ init ---- embedding unziped ------:", _EMBEDDING_PATH)



import uvicorn
from sse_starlette.sse import EventSourceResponse
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from loguru import logger
from langchain.embeddings import HuggingFaceEmbeddings
import os
import pathlib
import asyncio

import time
import json
from contextlib import asynccontextmanager
from concurrent.futures.thread import ThreadPoolExecutor
from chatglm_src.hf import HF
from chatglm_src.hf_server.models import ModelList, ModelCard, EmbedingResponse, MessageRequest, ls_all_can_loaed_models
from chatglm_src.hf_server.models import ModelCallRequest, ModelCallResponse
from chatglm_src.hf_server.models import NumpyEncoder
from chatglm_src.hf_server.models import ChatCompletionRequest,ChatCompletionResponseChoice,ChatCompletionResponse,ChatMessage
from chatglm_src.hf_server.utils import process_response, generate_chatglm3
from chatglm_src.hf_server.models import predict, UsageInfo, FunctionCallResponse
from chatglm_src.hf_server.models import MachineInfo,ClearRequest

try:
    import torch
    @asynccontextmanager
    async def lifespan(app: FastAPI):  # collects GPU memory
        yield
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.ipc_collect()

    app = FastAPI(lifespan=lifespan)
except ImportError:
    app = FastAPI()


try:
    TEXT_EMB_PATH = pathlib.Path("~").expanduser() / ".cache"/ "bge-large-zh"
    # TEXT_EMB_EN_PATH = pathlib.Path("~").expanduser() / ".cache"/ "bge-large-en"
    
    embeding_zh = HuggingFaceEmbeddings(model_name=str(TEXT_EMB_PATH))
    # embeding_en = HuggingFaceEmbeddings(model_name=str(TEXT_EMB_EN_PATH))
    embeding_en = None
except  Exception as e:
    print("Load Embedings Failed!!",e)
ThreadingExecutor = ThreadPoolExecutor(5)
LoadedModels  = {}

LLMModel = None
LLModelTokenizer = None

def load_llm(model_id) -> bool:
    global LLMModel, LLModelTokenizer
    MODEL_PATH = None
    for m in ls_all_can_loaed_models().data:
        if m.id == model_id:
            MODEL_PATH = m.path
            break
    if MODEL_PATH is not None:
        try:
            from transformers import AutoModel, AutoTokenizer
            from chatglm_src.hf_server.utils import load_model_on_gpus
            logger.info(f"Use GPU:{torch.cuda.device_count()}")
            LLMModel = load_model_on_gpus(MODEL_PATH, num_gpus=torch.cuda.device_count())
            LLModelTokenizer =  AutoTokenizer.from_pretrained(MODEL_PATH, trust_remote_code=True)
            return True
        except Exception as e:
            print(e)
    return False

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def loaded_models()-> ModelList:
    global LoadedModels
    res = []
    ss = ls_all_can_loaed_models()
    for i in LoadedModels:
        for s in ss.data:
            if s.id == i:
                res.append(s)
    return ModelList(data=res)


@app.get("/v1/models", response_model=ModelList)
async def list_models():
    return ls_all_can_loaed_models()



@app.post("/v1/models/load", response_model=ModelList)
async def load_model(request: ModelCard):
    global LoadedModels
    if request.id  not in LoadedModels:
        can = False
        for m in ls_all_can_loaed_models().data:
            if m.id == request.id:
                can = True
                break
        if can:
            HF.load_model(request.id)
            LoadedModels.update(HF.mos)
    return loaded_models()

async def call_model(model, docs):
    # global ThreadingExecutor
    loop = asyncio.get_running_loop()
    
    res = await loop.run_in_executor(None, model, docs)
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.ipc_collect()
    return res

async def call_stream(id,model,docs) -> ModelCallResponse:
    batch = []
    
    for i,doc in enumerate(docs):
        batch.append(doc)
        if len(batch) % 20 == 0 and len(batch) > 0:
            st = time.time()
            # res = model(batch)
            logger.info(f"batch before")
            res = await call_model(model, batch)
            et = time.time() - st
            logger.info(f"batch:{len(docs)} : {i}")
            try:
                e =  ModelCallResponse(id=id, data=res, used=et)
                reply = json.dumps(e.dict(exclude_unset=True), cls=NumpyEncoder) 
                yield "{}".format(reply)
                
                batch = []
            except Exception as e:
                print(type(res), type(res[0]),print(res[0]))
                logger.error(e)

    if len(batch) > 0:
        st = time.time()
        # res = model(batch)
        res = await call_model(model, batch)
        et = time.time() - st
        logger.info(f"batch:{len(docs)} : last")
        try:
            e =  ModelCallResponse(id=id, data=res, used=et)
            reply = json.dumps(e.dict(exclude_unset=True), cls=NumpyEncoder) 
            yield "{}".format(reply)

        except Exception as e:
            print(type(res), type(res[0]),print(res[0]))
            logger.error(e)


@app.post("/v1/models/call", response_model=ModelCallResponse)
async def call_smalle_models(request: ModelCallRequest):
    global LoadedModels
    if request.id  not in LoadedModels:
        can = False
        for m in ls_all_can_loaed_models().data:
            if m.id == request.id:
                can = True
                break
        if can:
            HF.load_model(request.id)
            LoadedModels.update(HF.mos)
    if request.id in LoadedModels:
        model = LoadedModels[request.id]
        if request.stream:
            logger.info(f"call : {request.id} streams")
            return EventSourceResponse(call_stream(request.id, model, request.messages))
        else:
            logger.info(f"call : {request.id} no streams")
            st = time.time()
            res = model(request.messages)
            rr = json.loads(json.dumps(res, cls=NumpyEncoder))
            return ModelCallResponse(id=request.id,data=rr, used=time.time()-st)
    else:
        return ModelCallResponse(id=request.id,data=[],error="no such model id:"+request.id, used=0.1)



async def embeding_stream(docs) -> EmbedingResponse:
    batch = []
    for i, doc in enumerate(docs):
        batch.append(doc)
        if len(batch) % 100 == 0 and len(batch) > 0:
            res_batch = []
            st = time.time()
            res_batch = await embeding_zh.aembed_documents(batch)
            et =  time.time() - st
            logger.info(f"batch:{len(docs)} : {i}")
            e =  EmbedingResponse(data=res_batch, len=len(res_batch), used=et)
            reply = json.dumps(e.dict(exclude_unset=True), cls=NumpyEncoder) 
            yield "{}".format(reply)
            batch = []
    
        

    if len(batch) > 0:
        logger.info(f"batch:{len(docs)} : last")
        st = time.time()
        res_batch = await embeding_zh.aembed_documents(batch)
        et = time.time() - st
        e =  EmbedingResponse(data=res_batch, len=len(res_batch), used=et)
        reply = json.dumps(e.dict(exclude_unset=True), cls=NumpyEncoder) 
        yield "{}".format(reply)
        
        
@app.get("/v1/device/info", response_model=MachineInfo)
async def machine_info():
    
    res = [i.split("|")[2].strip() for i in os.popen("nvidia-smi ").read().split("\n") if "Default " in i]
    g = []
    for i in res:
        if "/" in i:
            use, all = i.split("/",1)
            g.append({
                "used": int(use.strip()[:-3]),
                "all":  int(all.strip()[:-3]),
            })
    return MachineInfo(gpu=len(res), gpu_usage=g, loaded_models=loaded_models())

@app.post("/v1/models/clear", response_model=MachineInfo)
async def model_clear(request: ClearRequest):
    global LoadedModels
    for id in request.ids:
        if id in LoadedModels:
            m = LoadedModels[id]
            m.clean()
            del LoadedModels[id]

    res = [i.split("|")[2].strip() for i in os.popen("nvidia-smi ").read().split("\n") if "Default " in i]
    g = []
    for i in res:
        if "/" in i:
            use, all = i.split("/",1)
            g.append({
                "used": int(use.strip()[:-3]),
                "all":  int(all.strip()[:-3]),
            })
    return MachineInfo(gpu=len(res), gpu_usage=g, loaded_models=loaded_models())




@app.post("/v1/embeddings", response_model=EmbedingResponse)
async def embeding(request: MessageRequest):
    global embeding_zh ,embeding_en
    request.messages
    if request.stream:
        return  EventSourceResponse(embeding_stream(request.messages))
    st = time.time()
    res = embeding_zh.embed_documents(request.messages)
    et = time.time() - st
    return EmbedingResponse(data=res,len=len(res), used=et)


@app.post("/v1/chat/completions", response_model=ChatCompletionResponse)
async def create_chat_completion(request: ChatCompletionRequest):
    global LLMModel, LLModelTokenizer

    if LLMModel is None:
        if not load_llm(request.model):
            usage = UsageInfo()
            choice_data = ChatCompletionResponseChoice(
                index=0,
                message=ChatMessage(role="function", content=f"Err in Not found model {request.model} path"),
                finish_reason=finish_reason,
            )
            return ChatCompletionResponse(model=request.model, choices=[choice_data], object="chat.completion", usage=usage)
    model, tokenizer = LLMModel, LLModelTokenizer

    if len(request.messages) < 1 or request.messages[-1].role == "assistant":
        raise HTTPException(status_code=400, detail="Invalid request")

    gen_params = dict(
        messages=request.messages,
        temperature=request.temperature,
        top_p=request.top_p,
        max_tokens=request.max_tokens or 1024,
        echo=False,
        stream=request.stream,
        repetition_penalty=request.repetition_penalty,
        functions=request.functions,
    )

    logger.debug(f"==== request ====\n{gen_params}")

    if request.stream:
        generate = predict(model,tokenizer,request.model, gen_params)
        return EventSourceResponse(generate, media_type="text/event-stream")

    response = generate_chatglm3(model, tokenizer, gen_params)

    # Remove the first newline character
    if response["text"].startswith("\n"):
        response["text"] = response["text"][1:]
    response["text"] = response["text"].strip()
    usage = UsageInfo()
    function_call, finish_reason = None, "stop"
    if request.functions:
        try:
            function_call = process_response(response["text"], use_tool=True)
        except:
            logger.warning("Failed to parse tool call, maybe the response is not a tool call or have been answered.")

    if isinstance(function_call, dict):
        finish_reason = "function_call"
        function_call = FunctionCallResponse(**function_call)

    message = ChatMessage(
        role="assistant",
        content=response["text"],
        function_call=function_call if isinstance(function_call, FunctionCallResponse) else None,
    )

    logger.debug(f"==== message ====\n{message}")

    choice_data = ChatCompletionResponseChoice(
        index=0,
        message=message,
        finish_reason=finish_reason,
    )
    task_usage = UsageInfo.model_validate(response["usage"])
    for usage_key, usage_value in task_usage.model_dump().items():
        setattr(usage, usage_key, getattr(usage, usage_key) + usage_value)
    return ChatCompletionResponse(model=request.model, choices=[choice_data], object="chat.completion", usage=usage)


def run(port=15001, workers=1):
    uvicorn.run(app, host='0.0.0.0', port=port, workers=workers)

if __name__ == "__main__":
    run()



import pathlib
from typing import Any
from termcolor import colored
import datetime

from .qwen import VllmBase
from langchain.llms.base import LLM
from hashlib import md5
from aiowebsocket.converses import AioWebSocket
from websocket import create_connection
import websockets
from websockets.server import serve
import time
import asyncio
import json
import numpy as np

try:
    import torch
    from transformers import AutoModelForCausalLM, AutoTokenizer
    from transformers.generation.utils import GenerationConfig
except Exception as e:
    pass



def load_model_on_gpus_old(checkpoint_path, num_gpus: int = 2,device_map = "auto"):
    tokenizer = AutoTokenizer.from_pretrained(checkpoint_path, use_fast=False, trust_remote_code=True)
    model = AutoModelForCausalLM.from_pretrained(checkpoint_path, device_map=device_map, num_gpus=num_gpus, torch_dtype=torch.bfloat16, trust_remote_code=True)
    model.generation_config = GenerationConfig.from_pretrained(checkpoint_path)


    return model,tokenizer

class Baichuan2LLM(LLM):
    max_token: int = 4096
    temperature: float = 0.01
    top_p = 0.9
    history = []
    history_id = "default"
    tokenizer: Any = None
    model: Any = None
    history_len: int = 10
    model: Any = None
    tokenizer: Any = None
    cpu: bool = False
    streaming: bool = False
    verbose: bool = False
    callbacks :Any  = []
    # callback_manager:  Any = BaseCallbackManager
    remote_host: Any = None
    tokenizer: Any = None
    model: Any = None
    history_len: int = 10
    model: Any = None
    tokenizer: Any = None
    model_path: str = pathlib.Path.home() / ".cache" / "baichuan2"
    streaming: bool = False
    remote_host: str = None
    
    @classmethod
    def load(cls, *args,model_path=None, **kargs):
        mo = cls(*args, **kargs)
        if model_path is not None:
            mo.model_path = pathlib.Path(model_path)
        
        if mo.model_path.exists() and mo.remote_host is None:
            if torch.cuda.device_count() > 0:
                print(colored("[GPU]","green"),":",torch.cuda.device_count(), "GPUs" )
                # use baichuan2
                try:
                    mo.model, mo.tokenizer = load_model_on_gpus_old(mo.model_path, num_gpus=torch.cuda.device_count())

                except Exception as e:
                    print(colored("[GPU]","yellow"),": failed !! but try load use new way " + str(e),torch.cuda.device_count(), "GPUs" )
        else:
            pass
        return mo        

    def make_args(self, parameter):
        # parameter["stop"] = ['<|im_end|>','<|endoftext|>']
        # parameter["stop"] = '<|im_end>'
        return parameter

    def make_context(self, prompt, history=[],system="You are a assistant."):
        def _token(role, content):
            return f"{role}:\n{content}"

        raw_text = _token("system", system) + "\n"
        for q, h in history:
            raw_text += _token("user", q)
            raw_text += "\n"+_token("assistant", h) +"\n"
        return raw_text + _token("user", prompt) + "\n" + self.start_token + "assistant\n"

    def parse_output(self, output):
        return output
    
    async def asend_to_remote(self, data, ws):
        """
        every chunk of data is 2M
        """
        for i in range(0, len(data), 1024*1024*2):
            await ws.send(data[i:i+1024*1024*2])
        await ws.send("[STOP]")
    
    async def aclient_send(self, prompt):
        
        uri = f"ws://{self.remote_host}:15000"
        result = ''
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        # self.callback_manager =  BaseCallbackManager(self.callbacks)
        # self.callback_manager.set_handlers(self.callbacks)
        async with AioWebSocket(uri) as aws:
            converse = aws.manipulator
            
            user_id = md5(time.asctime().encode()).hexdigest()
            await converse.send(json.dumps({"user_id":user_id, "password":PASSWORD}).encode())
            res = await converse.receive()
            res = res.decode()
            if res != "ok":
                raise Exception("password error:"+res)
            data = json.dumps({"prompt":prompt, "history":self.history,"model":"chatglm","temperature":self.temperature}).encode()
            await self.asend_to_remote(data, converse)

            # for call in self.callbacks:
            #     if is_async_method(call.on_llm_start):
            #         await call.on_llm_start(prompt, None, verbose=self.verbose)
            while 1:
                res = await converse.receive()
                msg = json.loads(res.decode())
                # { "new":delta,"response": response, "history": history,"query": prompt}
                if "stop" in msg:
                    break
                new_token = msg["new"]
                response = msg["response"]
                msg["verbose"] = self.verbose
                # for call in self.callbacks:
                #     if is_async_method(call.on_llm_start):
                #         await call.on_llm_new_token(new_token, **msg)
                yield msg
                result = response
        self.history = self.history+[[prompt, result]]
        # for call in self.callbacks:
        #     if is_async_method(call.on_llm_start):
        #         await call.on_llm_end(result, verbose=self.verbose)
        # return result

            

    
    def client_send(self, prompt):
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        assert self.remote_host is not None
        uri = f"ws://{self.remote_host}:15000"
        result = ''
        # self.callback_manager =  BaseCallbackManager(self.callbacks)
        # self.callback_manager.set_handlers(self.callbacks)
        
        ws = create_connection(f"ws://{self.remote_host}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        # self.callback_manager =  langchain.callbacks.base.BaseCallbackManager(self.callbacks)
        
        # self.callback_manager =  BaseCallbackManager(self.callbacks)
        # self.callback_manager.set_handlers(self.callbacks)
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        # time.sleep(0.5)
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        result = ''
            
            
        data = json.dumps({"prompt":prompt, "history":self.history,"model":"baichuan2-7b", "temperature": self.temperature})
        self.send_to_remote(data, ws)
        while 1:
            res = ws.recv()
            try:
                msg = json.loads(res)
            except Exception as e:
                print(colored("[info]:","yellow") ,res)
                continue
            # { "new":delta,"response": response, "history": history,"query": prompt}
            if "stop" in msg:
                break
            # new_token = msg["new"]
            # response = msg["response"]
            # history = msg["history"]
            msg["verbose"] = self.verbose
            result = msg
            yield msg
            
            # for call in self.callbacks:
            #     if is_async_method(call.on_llm_start):
                    # await call.on_llm_end(result, verbose=self.verbose)
        self.history = self.history+[[prompt, result]]
        # yield result
    

    def send_to_remote(self,data,ws):
        """
        every chunk of data is 2M
        """
        for i in range(0, len(data), 1024*1024*2):
            ws.send(data[i:i+1024*1024*2])
        ws.send("[STOP]")

    @property
    def _llm_type(self) -> str:
        return "baichuan2-7b"

    def _call(self, prompt):
        for data in self.stream(prompt):
            delta = data["new"]
            self.callback_manager.on_llm_new_token(
                    delta, verbose=self.verbose, **data
                )
        return data["response"]
    
    async def _acall(self, prompt):
        async for data in self.aclient_send(prompt):
            delta = data["new"]
            self.callback_manager.on_llm_new_token(
                delta, verbose=self.verbose, **data
            )
        return data["response"]
    
    def stream(self, prompt):
        if self.remote_host is not None:
            for result in self.client_send(prompt):
                yield result
            
        else:
            position = 0
            messages = []
            for u,h in self.history:
                messages.append({"role":"user", "content":u})
                messages.append({"role":"assistant", "content":h})
            messages.append({"role":"user", "content":prompt})
                
            for response in self.model.chat(self.tokenizer, messages, stream=True):
                delta = response[position:]
                position = len(response)
                data = {"new": delta, "response": response, "history": self.history, "query": prompt}
                if torch.backends.mps.is_available():
                    torch.mps.empty_cache()
                yield data
            self.history.append((prompt, response))



import datetime
import time
import json
import inspect

from typing import Iterable, List
from aiowebsocket.converses import AioWebSocket
from websocket import create_connection
from hashlib import md5
from termcolor import colored

def is_async_method(method):
    return inspect.iscoroutinefunction(method)

class BaseLLM:
    remote_host :str = None
    callbacks : List = None

    def send_to_remote(self,data,ws):
        """
        every chunk of data is 2M
        """
        for i in range(0, len(data), 1024*1024*2):
            ws.send(data[i:i+1024*1024*2])
        ws.send("[STOP]")

    async def asend_to_remote(self, data, ws):
        """
        every chunk of data is 2M
        """
        for i in range(0, len(data), 1024*1024*2):
            await ws.send(data[i:i+1024*1024*2])
        await ws.send("[STOP]")

    def remote_call(self, prompt, stop):
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        assert self.remote_host is not None
        
        ws = create_connection(f"ws://{self.remote_host}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        # self.callback_manager =  langchain.callbacks.base.BaseCallbackManager(self.callbacks)
        
        # self.callback_manager =  BaseCallbackManager(self.callbacks)
        # self.callback_manager.set_handlers(self.callbacks)
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        # time.sleep(0.5)
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        result = ''
        data = json.dumps({
            "model":"vicuna-13b",
            "prompt":prompt,
            "history":self.history,
            "temperature": self.temperature,
            })
        self.send_to_remote(data, ws)
        
        for callback in self.callbacks:
            if  is_async_method(callback.on_llm_start):continue
            callback.on_llm_start(
                None,
                prompt,
                run_id=user_id,
                verbose=self.verbose
            )
        while 1:
            res = ws.recv()
            msg = json.loads(res)
            # { "new":delta,"response": response, "history": history,"query": prompt}
            if "stop" in msg:
                break
            new_token = msg["new"]
            response = msg["response"]
            history = msg["history"]
            msg["verbose"] = self.verbose
            # self.remote_callback(new_token, history, response)
            
            # self.callback_manager.on_llm_new_token(new_token, **msg)
            for callback in self.callbacks:
                if  is_async_method(callback.on_llm_new_token):continue
                callback.on_llm_new_token(
                    new_token,
                    **msg
                )
            result = response
        for callback in self.callbacks:
            if  is_async_method(callback.on_llm_end):continue
            callback.on_llm_end(result, verbose=self.verbose)
        
        self.history = self.history+[[prompt, result]]
        return result

    async def async_remote_call(self, prompt):
        uri = f"ws://{self.remote_host}:15000"
        result = ''
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        # self.callback_manager =  BaseCallbackManager(self.callbacks)
        # self.callback_manager.set_handlers(self.callbacks)
        async with AioWebSocket(uri) as aws:
            converse = aws.manipulator
            
            user_id = md5(time.asctime().encode()).hexdigest()
            await converse.send(json.dumps({"user_id":user_id, "password":PASSWORD}).encode())
            res = await converse.receive()
            res = res.decode()
            if res != "ok":
                raise Exception("password error:"+res)
            data = json.dumps({"prompt":prompt, "history":self.history,"model":"chatglm","temperature":self.temperature}).encode()
            await self.asend_to_remote(data, converse)

            for call in self.callbacks:
                if is_async_method(call.on_llm_start):
                    await call.on_llm_start(prompt, None, verbose=self.verbose)
            while 1:
                res = await converse.receive()
                msg = json.loads(res.decode())
                # { "new":delta,"response": response, "history": history,"query": prompt}
                if "stop" in msg:
                    break
                new_token = msg["new"]
                response = msg["response"]
                msg["verbose"] = self.verbose
                for call in self.callbacks:
                    if is_async_method(call.on_llm_start):
                        await call.on_llm_new_token(new_token, **msg)
                result = response
        self.history = self.history+[[prompt, result]]
        for call in self.callbacks:
            if is_async_method(call.on_llm_start):
                await call.on_llm_end(result, verbose=self.verbose)
        return result
    

    def stream(self,prompt: str, stop: List[str] = None):
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        assert self.remote_host is not None
        uri = f"ws://{self.remote_host}:15000"
        result = ''
        # self.callback_manager =  BaseCallbackManager(self.callbacks)
        # self.callback_manager.set_handlers(self.callbacks)
        
        ws = create_connection(f"ws://{self.remote_host}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        # self.callback_manager =  langchain.callbacks.base.BaseCallbackManager(self.callbacks)
        
        # self.callback_manager =  BaseCallbackManager(self.callbacks)
        # self.callback_manager.set_handlers(self.callbacks)
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        # time.sleep(0.5)
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        result = ''
            
            
        data = json.dumps({"prompt":prompt, "history":self.history,"model":"chatglm", "temperature": self.temperature})
        self.send_to_remote(data, ws)
        while 1:
            res = ws.recv()
            msg = json.loads(res)
            # { "new":delta,"response": response, "history": history,"query": prompt}
            if "stop" in msg:
                break
            new_token = msg["new"]
            response = msg["response"]
            history = msg["history"]
            msg["verbose"] = self.verbose
            result = msg
            yield msg
            
            # for call in self.callbacks:
            #     if is_async_method(call.on_llm_start):
                    # await call.on_llm_end(result, verbose=self.verbose)
        self.history = self.history+[[prompt, result]]


from typing import List, Literal, Optional, Union, Any,Dict
from pydantic import BaseModel, Field
import time
import os
import pathlib
import json
import numpy as np
from loguru import logger
from .utils import generate_stream_chatglm3, process_response
class ModelCard(BaseModel):
    id: str
    object: str = "model"
    created: int = Field(default_factory=lambda: int(time.time()))
    owned_by: str = "owner"
    root: Optional[str] = None
    parent: Optional[str] = None
    path: Optional[str] = None
    permission: Optional[list] = None


class ModelList(BaseModel):
    object: str = "list"
    data: List[ModelCard] = []


class FunctionCallResponse(BaseModel):
    name: Optional[str] = None
    arguments: Optional[str] = None


class MessageRequest(BaseModel):
    messages: List[str]
    stream: Optional[bool] = False

class EmbedingResponse(BaseModel):
    len: Optional[int]
    id: str = "embeding_zh"
    data: Any
    used: float


class ModelCallRequest(BaseModel):
    id: str
    messages: List[str]
    stream: Optional[bool]
    

class ModelCallResponse(BaseModel):
    id: str
    data: Union[dict, List[List[dict]],List[dict], List[str], List[int]]
    used: float
    error: Optional[str]



class ChatMessage(BaseModel):
    role: Literal["user", "assistant", "system", "function"]
    content: str = None
    name: Optional[str] = None
    function_call: Optional[FunctionCallResponse] = None


class DeltaMessage(BaseModel):
    role: Optional[Literal["user", "assistant", "system"]] = None
    content: Optional[str] = None
    function_call: Optional[FunctionCallResponse] = None


class ChatCompletionRequest(BaseModel):
    model: str
    messages: List[ChatMessage]
    temperature: Optional[float] = 0.8
    top_p: Optional[float] = 0.8
    max_tokens: Optional[int] = None
    stream: Optional[bool] = False
    functions: Optional[Union[dict, List[dict]]] = None
    # Additional parameters
    repetition_penalty: Optional[float] = 1.1


class ChatCompletionResponseChoice(BaseModel):
    index: int
    message: ChatMessage
    finish_reason: Literal["stop", "length", "function_call"]


class ChatCompletionResponseStreamChoice(BaseModel):
    index: int
    delta: DeltaMessage
    finish_reason: Optional[Literal["stop", "length", "function_call"]]


class UsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0
    completion_tokens: Optional[int] = 0

class MachineInfo(BaseModel):
    gpu: int = 0
    memory_use: Optional[float]
    gpu_usage: Optional[List[Dict]]
    loaded_models: Optional[ModelList]

class ChatCompletionResponse(BaseModel):
    model: str
    object: Literal["chat.completion", "chat.completion.chunk"]
    choices: List[Union[ChatCompletionResponseChoice, ChatCompletionResponseStreamChoice]]
    created: Optional[int] = Field(default_factory=lambda: int(time.time()))
    usage: Optional[UsageInfo] = None

class ClearRequest(BaseModel):
    ids: List[str]

async def predict(model, tokenizer, model_id: str, params: dict):
    

    choice_data = ChatCompletionResponseStreamChoice(
        index=0,
        delta=DeltaMessage(role="assistant"),
        finish_reason=None
    )
    chunk = ChatCompletionResponse(model=model_id, choices=[choice_data], object="chat.completion.chunk")
    
    yield "{}".format(chunk.json(exclude_unset=True))

    previous_text = ""
    for new_response in generate_stream_chatglm3(model, tokenizer, params):
        decoded_unicode = new_response["text"]
        delta_text = decoded_unicode[len(previous_text):]
        previous_text = decoded_unicode

        finish_reason = new_response["finish_reason"]
        if len(delta_text) == 0 and finish_reason != "function_call":
            continue

        function_call = None
        if finish_reason == "function_call":
            try:
                function_call = process_response(decoded_unicode, use_tool=True)
            except:
                logger.warning("Failed to parse tool call, maybe the response is not a tool call or have been answered.")

        if isinstance(function_call, dict):
            function_call = FunctionCallResponse(**function_call)

        delta = DeltaMessage(
            content=delta_text,
            role="assistant",
            function_call=function_call if isinstance(function_call, FunctionCallResponse) else None,
        )

        choice_data = ChatCompletionResponseStreamChoice(
            index=0,
            delta=delta,
            finish_reason=finish_reason
        )
        chunk = ChatCompletionResponse(model=model_id, choices=[choice_data], object="chat.completion.chunk")
        yield "{}".format(chunk.json(exclude_unset=True))

    choice_data = ChatCompletionResponseStreamChoice(
        index=0,
        delta=DeltaMessage(),
        finish_reason="stop"
    )
    chunk = ChatCompletionResponse(model=model_id, choices=[choice_data], object="chat.completion.chunk")
    yield "{}".format(chunk.json(exclude_unset=True))
    yield '[DONE]'





def ls_all_can_loaed_models():
    models = []
    # model_card = ModelCard(id="chatglm3-6b")
    e = pathlib.Path("~/.cache/").expanduser()
    for n in  os.listdir(e):
        if n.endswith("/"):
            n = n[:-1]
        repo =  e / n / "config.json"
        if repo.exists():
            models.append(ModelCard(id=n, path=str(e/n)))
    for n in os.listdir( e / "THUDM/" ):
        if n.endswith("/"):
            n = n[:-1]
        repo =  e / "THUDM"/ n / "config.json"
        if repo.exists():
            models.append(ModelCard(id=n, path=str(e/"THUDM"/n) ))
    return ModelList(data=models)

class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)
    


import requests
import aiohttp
import tqdm
from .models import *


def R(uri, object:BaseModel=None, method='post',use_stream=False, **datas):
    data = {}
    data.update(datas)
    o = None
    if object is not None:
        o = object.dict()
        if "stream" in o:
            # print("Set Stream:",use_stream)
            object.stream = use_stream
        o = object.dict()
        data.update(o)
        
    if method == "post":
        response = requests.post(uri, json=data, stream=use_stream)
    else:
        response = requests.get(uri, json=data, stream=use_stream)
    if response.status_code == 200:
        if use_stream:
            # 处理流式响应
            if "messages" in o:
                T = (len(object.messages) // 20) + 1
                if uri.endswith("embeddings"):
                    T = (len(object.messages) // 100) + 1
                bar = tqdm.tqdm(total=T,desc=" + deal data")
                bar.leave = False
                for line in response.iter_lines():
                    if line:
                        if line[:6] == b": ping" and line[6] != b"{":continue
                        decoded_line = line.decode('utf-8')[6:].strip()
                        if decoded_line:
                            try:
                                response_json = json.loads(decoded_line)
                                bar.update(1)
                                yield response_json
                            except Exception as e:
                                print("Special Token:", e, len(line), line)
                bar.clear()
                bar.close()
        else:
            # 处理非流式响应
            decoded_line = response.json()
            yield decoded_line
    else:
        print("Error:", response.status_code)
        return None



def create_chat_completion(url, messages:List[ChatMessage], functions=None, use_stream=True,model="chatglm3-6b-32k",temperature=0.8,top_p=0.8, max_tokens=8000):
    # data = {
    #     "function": functions,  # 函数定义
    #     "model": model,  # 模型名称
    #     "messages": messages,  # 会话历史
    #     "stream": use_stream,  # 是否流式响应
    #     "max_tokens": max_tokens,  # 最多生成字数
    #     "temperature": temperature,  # 温度
    #     "top_p": top_p,  # 采样概率
    # }
    req = ChatCompletionRequest(
        model=model, 
        functions=functions,
        messages=messages,
        top_p=top_p,
        temperature=temperature,
        stream=use_stream,
        max_tokens=max_tokens,
    )
    data = req.dict()
    response = requests.post(url, json=data, stream=use_stream)
    if response.status_code == 200:
        if use_stream:
            # 处理流式响应
            for line in response.iter_lines():
                if line:
                    decoded_line = line.decode('utf-8')[6:]
                    try:
                        if decoded_line.strip():
                            response_json = json.loads(decoded_line.strip())
                            # content = response_json.get("choices", [{}])[0].get("delta", {}).get("content", "")
                            yield ChatCompletionResponse.parse_obj(response_json)
                            
                    except:
                        # print("Special Token:", decoded_line)
                        pass
        else:
            # 处理非流式响应
            decoded_line = response.json()
            # content = decoded_line.get("choices", [{}])[0].get("message", "").get("content", "")
            yield ChatCompletionResponse.parse_obj(decoded_line)
    else:
        print("Error:", response.status_code)
        return None

async def acreate_chat_completion(url, messages:List[ChatMessage], functions=None, use_stream=True,model="chatglm3-6b-32k",temperature=0.8,top_p=0.8, max_tokens=8000):
    req = ChatCompletionRequest(
        model=model, 
        functions=functions,
        messages=messages,
        top_p=top_p,
        temperature=temperature,
        stream=use_stream,
        max_tokens=max_tokens,
    )
    data = req.dict()
    # print()
    async with aiohttp.ClientSession() as session:
        async with session.post(url, json=data) as response:
    
            if response.status == 200:
                if use_stream:
                    # 处理流式响应
                    # import ipdb;ipdb.set_trace()
                    while 1:
                        try:
                            line = await response.content.readline()
                            if line:
                                decoded_line = line.decode('utf-8')[6:]
                                try:
                                    if decoded_line.strip():
                                        response_json = json.loads(decoded_line.strip())
                                        # content = response_json.get("choices", [{}])[0].get("delta", {}).get("content", "")
                                        c = ChatCompletionResponse.parse_obj(response_json)
                                        
                                        if c.choices[-1].finish_reason == "stop":
                                            break
                                        yield c
                                        
                                        
                                except:
                                    print("Special Token:", decoded_line)
                        except EOFError:
                            break
                    
                        
                else:
                    # 处理非流式响应
                    decoded_line = response.json()
                    # content = decoded_line.get("choices", [{}])[0].get("message", "").get("content", "")
                    yield ChatCompletionResponse.parse_obj(decoded_line)
            else:
                print("Error:", response.status)
                

def simple_chat(url, messages,use_stream=True):
    functions = None
    chat_messages = [
        {
            "role": "system",
            "content": "You are ChatGLM3, a helpful assistant. Follow the user's instructions carefully. Respond using markdown.",
        },
        
    ]
    for message in messages:
        chat_messages.append(
            {
                "role": "user",
                "content": message
            }
        )
    return create_chat_completion( messages=chat_messages, functions=functions, use_stream=use_stream)
